from copy import deepcopy

import numpy as np
import pytest

from luxonis_ml.data import AlbumentationsEngine
from luxonis_ml.typing import Labels


@pytest.fixture(
    params=[
        {"image": np.zeros((320, 320, 3), dtype=np.uint8)},
        {
            "rgb_image": np.zeros((320, 320, 3), dtype=np.uint8),
            "ir_image": np.zeros((320, 320, 1), dtype=np.uint8),
        },
        {
            "left_img": np.zeros((320, 320, 1), dtype=np.uint8),
            "right_img": np.zeros((320, 320, 1), dtype=np.uint8),
            "middle_img": np.zeros((320, 320, 3), dtype=np.uint8),
        },
    ]
)
def images_dict(request: pytest.FixtureRequest) -> dict[str, np.ndarray]:
    return request.param


@pytest.fixture
def labels() -> Labels:
    return {
        "task/classification": np.array([1.0]),
        "task/boundingbox": np.array(
            [
                [0.0, 0.57, 0.30, 0.17, 0.25],
                [0.0, 0.39, 0.27, 0.20, 0.10],
            ]
        ),
        "task/keypoints": np.array(
            [
                [0.69, 0.37, 2.0, 0.0, 0.0, 0.0, 0.68, 0.36, 2.0],
                [0.51, 0.33, 2.0, 0.0, 0.0, 0.0, 0.50, 0.32, 2.0],
            ]
        ),
        "task/segmentation": np.zeros((1, 320, 320)),
    }


@pytest.fixture
def targets() -> dict[str, str]:
    return {
        "task/boundingbox": "boundingbox",
        "task/keypoints": "keypoints",
        "task/segmentation": "segmentation",
    }


@pytest.fixture
def n_classes() -> dict[str, int]:
    return {
        "task/boundingbox": 1,
        "task/keypoints": 1,
        "task/segmentation": 1,
    }


def test_mosaic4(
    images_dict: dict[str, np.ndarray],
    labels: Labels,
    targets: dict[str, str],
    n_classes: dict[str, int],
) -> None:
    config = [
        {
            "name": "Mosaic4",
            "params": {"p": 1.0, "out_width": 640, "out_height": 640},
        }
    ]
    source_names = list(images_dict.keys())
    augmentations = AlbumentationsEngine(
        256, 256, targets, n_classes, source_names, config
    )
    augmentations.apply([(images_dict, deepcopy(labels)) for _ in range(4)])


def test_mixup(
    images_dict: dict[str, np.ndarray],
    labels: Labels,
    targets: dict[str, str],
    n_classes: dict[str, int],
) -> None:
    config = [{"name": "MixUp", "params": {"p": 1.0}}]
    source_names = list(images_dict.keys())
    augmentations = AlbumentationsEngine(
        256, 256, targets, n_classes, source_names, config
    )
    augmentations.apply([(images_dict, deepcopy(labels)) for _ in range(2)])


def test_at_least_one_bbox_random_crop() -> None:
    """Test that AtLeastOneBBoxRandomCrop guarantees at least one bbox.

    This is a test for ensuring that the correct "bboxes" key is passed
    to Albumentations transforms that read data["bboxes"] directly.
    """
    image = np.random.randint(0, 255, (320, 320, 3), dtype=np.uint8)
    images_dict = {"image": image}
    labels: Labels = {
        "task/boundingbox": np.array(
            [
                [0.0, 0.5, 0.5, 0.1, 0.1],
            ]
        ),
    }
    targets = {"task/boundingbox": "boundingbox"}
    n_classes = {"task/boundingbox": 1}
    config = [
        {
            "name": "AtLeastOneBBoxRandomCrop",
            "params": {
                "height": 40,
                "width": 40,
                "erosion_factor": 0.0,
                "p": 1.0,
            },
        }
    ]
    engine = AlbumentationsEngine(
        256, 256, targets, n_classes, ["image"], config
    )
    for _ in range(10):
        _, out_labels = engine.apply([(images_dict, deepcopy(labels))])
        bboxes = out_labels.get("task/boundingbox")
        assert bboxes is not None, (
            "AtLeastOneBBoxRandomCrop should produce bounding box output"
        )
        assert len(bboxes) > 0, (
            "AtLeastOneBBoxRandomCrop should guarantee at least one "
            "bounding box per crop"
        )


def test_batched_p_0(
    images_dict: dict[str, np.ndarray],
    labels: Labels,
    targets: dict[str, str],
    n_classes: dict[str, int],
) -> None:
    config = [
        {
            "name": "Mosaic4",
            "params": {"p": 0, "out_width": 640, "out_height": 640},
        },
        {"name": "MixUp", "params": {"p": 0}},
    ]
    source_names = list(images_dict.keys())
    augmentations = AlbumentationsEngine(
        256, 256, targets, n_classes, source_names, config
    )
    augmentations.apply([(images_dict, deepcopy(labels)) for _ in range(8)])



import albumentations as A

from luxonis_ml.utils import Registry

from .letterbox_resize import LetterboxResize
from .mixup import MixUp
from .mosaic import Mosaic4
from .symetric_keypoints_flip import (
    HorizontalSymetricKeypointsFlip,
    TransposeSymmetricKeypoints,
    VerticalSymetricKeypointsFlip,
)

TRANSFORMATIONS: Registry[type[A.BasicTransform]] = Registry(
    "albumentations_transformations"
)

TRANSFORMATIONS.register(module=LetterboxResize)
TRANSFORMATIONS.register(module=MixUp)
TRANSFORMATIONS.register(module=Mosaic4)
TRANSFORMATIONS.register(module=HorizontalSymetricKeypointsFlip)
TRANSFORMATIONS.register(module=VerticalSymetricKeypointsFlip)
TRANSFORMATIONS.register(module=TransposeSymmetricKeypoints)

__all__ = [
    "TRANSFORMATIONS",
    "HorizontalSymetricKeypointsFlip",
    "LetterboxResize",
    "MixUp",
    "Mosaic4",
    "TransposeSymmetricKeypoints",
    "VerticalSymetricKeypointsFlip",
]



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, cast

from luxonis_ml.data.exporters.base_exporter import BaseExporter
from luxonis_ml.data.exporters.exporter_utils import (
    PreparedLDF,
    check_group_file_correspondence,
    exporter_specific_annotation_warning,
    split_of_group,
)
from luxonis_ml.enums import DatasetType


class DarknetExporter(BaseExporter):
    def __init__(
        self,
        dataset_identifier: str,
        output_path: Path,
        max_partition_size_gb: float | None,
    ):
        super().__init__(
            dataset_identifier, output_path, max_partition_size_gb
        )
        self.class_to_id: dict[str, int] = {}
        self.class_names: list[str] = []

    @staticmethod
    def get_split_names() -> dict[str, str]:
        return {"train": "train", "val": "valid", "test": "test"}

    def supported_ann_types(self) -> list[str]:
        return DatasetType.DARKNET.supported_annotation_formats

    def export(self, prepared_ldf: PreparedLDF) -> None:
        check_group_file_correspondence(prepared_ldf)
        exporter_specific_annotation_warning(
            prepared_ldf, self.supported_ann_types()
        )

        labels_by_split: dict[str, dict[str, list[str]]] = {
            v: {} for v in self.get_split_names().values()
        }

        df = prepared_ldf.processed_df
        grouped = df.group_by(["file", "group_id"], maintain_order=True)

        copied_files: set[Path] = set()

        for key, group_df in grouped:
            file_name, group_id = cast(tuple[str, Any], key)
            split_key = split_of_group(prepared_ldf, group_id)
            split_name = self.get_split_names()[split_key]

            file_path = Path(str(file_name))
            idx = self.image_indices.setdefault(
                file_path, len(self.image_indices)
            )
            new_name = f"{idx}{file_path.suffix}"
            new_stem = Path(new_name).stem

            label_lines = self._collect_darknet_bounding_box_labels(group_df)

            ann_size = sum(len(line) for line in label_lines)
            img_size = file_path.stat().st_size
            labels_by_split = self._maybe_roll_partition(
                labels_by_split, ann_size + img_size
            )

            labels_by_split[split_name][new_stem] = label_lines

            split_dir = self._get_data_path(
                self.output_path, split_name, self.part
            )
            split_dir.mkdir(parents=True, exist_ok=True)
            dest_img = split_dir / new_name

            if file_path not in copied_files:
                copied_files.add(file_path)
                if dest_img != file_path:
                    dest_img.write_bytes(file_path.read_bytes())
                self.current_size += img_size

        self._dump_annotations(labels_by_split, self.output_path, self.part)

    def _collect_darknet_bounding_box_labels(
        self,
        group_df: Any,
    ) -> list[str]:
        label_lines: list[str] = []

        for row in group_df.iter_rows(named=True):
            ttype = row["task_type"]
            ann_str = row["annotation"]
            cname = row["class_name"]

            if ttype != "boundingbox" or ann_str is None:
                continue

            # Register class if new
            if cname and cname not in self.class_to_id:
                self.class_to_id[cname] = len(self.class_to_id)
                self.class_names.append(cname)

            if not cname or cname not in self.class_to_id:
                continue

            data = json.loads(ann_str)
            x_tl = float(data.get("x", 0.0))
            y_tl = float(data.get("y", 0.0))
            w = float(data.get("w", 0.0))
            h = float(data.get("h", 0.0))

            cx = x_tl + w / 2.0
            cy = y_tl + h / 2.0

            cid = self.class_to_id[cname]
            label_lines.append(f"{cid} {cx:.12f} {cy:.12f} {w:.12f} {h:.12f}")

        return label_lines

    def _maybe_roll_partition(
        self,
        labels_by_split: dict[str, dict[str, list[str]]],
        additional_size: int,
    ) -> dict[str, dict[str, list[str]]]:
        if (
            self.max_partition_size
            and self.part is not None
            and (self.current_size + additional_size) > self.max_partition_size
        ):
            self._dump_annotations(
                labels_by_split, self.output_path, self.part
            )
            self.current_size = 0
            self.part += 1
            return {v: {} for v in self.get_split_names().values()}
        return labels_by_split

    def _dump_annotations(
        self,
        labels_by_split: dict[str, dict[str, list[str]]],
        output_path: Path,
        part: int | None = None,
    ) -> None:
        if all(len(imgs) == 0 for imgs in labels_by_split.values()):
            return

        base = (
            output_path / f"{self.dataset_identifier}_part{part}"
            if part is not None
            else output_path / self.dataset_identifier
        )
        base.mkdir(parents=True, exist_ok=True)

        for split_name, per_image in labels_by_split.items():
            split_dir = base / split_name
            split_dir.mkdir(parents=True, exist_ok=True)

            for stem, lines in per_image.items():
                (split_dir / f"{stem}.txt").write_text(
                    "\n".join(lines) + ("\n" if lines else ""),
                    encoding="utf-8",
                )

            (split_dir / "_darknet.labels").write_text(
                "\n".join(self.class_names)
                + ("\n" if self.class_names else ""),
                encoding="utf-8",
            )

    def _get_data_path(
        self, output_path: Path, split_name: str, part: int | None = None
    ) -> Path:
        base = (
            output_path / f"{self.dataset_identifier}_part{part}"
            if part is not None
            else output_path / self.dataset_identifier
        )
        return base / split_name



from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import Any, cast

import pycocotools.mask as maskUtils
from loguru import logger
from PIL import Image

from luxonis_ml.data.exporters.base_exporter import BaseExporter
from luxonis_ml.data.exporters.exporter_utils import (
    PreparedLDF,
    check_group_file_correspondence,
    exporter_specific_annotation_warning,
    get_single_skeleton,
    split_of_group,
)
from luxonis_ml.data.utils import COCOFormat
from luxonis_ml.enums import DatasetType


class CocoExporter(BaseExporter):
    """Exporter for COCO dataset format (Roboflow/FiftyOne variants)."""

    def __init__(
        self,
        dataset_identifier: str,
        output_path: Path,
        max_partition_size_gb: float | None,
        format: COCOFormat = COCOFormat.ROBOFLOW,
        *,
        skeletons: dict[str, Any] | None = None,
    ):
        super().__init__(
            dataset_identifier, output_path, max_partition_size_gb
        )
        self.format = format
        self.skeletons = skeletons
        if self.skeletons is None:
            self.allow_keypoints = False
        elif len(self.skeletons) == 1:
            self.allow_keypoints = True
        else:
            self.allow_keypoints = False
            logger.warning(
                "Skipping keypoint annotations because COCO only supports a single keypoint export class."
                "To export multiple keypoint classes please use the Luxonis native export format"
            )

        splits = self.get_split_names()
        self.class_name_to_category_id: dict[str, dict[str, int]] = {
            s: {} for s in splits
        }
        self.last_category_id: dict[str, int] = dict.fromkeys(splits, 1)
        self.image_registry: dict[str, dict[str, dict[str, Any]]] = {
            s: {} for s in splits
        }

    def get_split_names(self) -> dict[str, str]:
        if self.format == COCOFormat.ROBOFLOW:
            return {"train": "train", "val": "valid", "test": "test"}
        return {"train": "train", "val": "validation", "test": "test"}

    def supported_ann_types(self) -> list[str]:
        return DatasetType.COCO.supported_annotation_formats

    def export(self, prepared_ldf: PreparedLDF) -> None:
        check_group_file_correspondence(prepared_ldf)
        exporter_specific_annotation_warning(
            prepared_ldf, self.supported_ann_types()
        )

        splits = self.get_split_names()
        annotation_splits: dict[str, dict[str, Any]] = {
            s: {"images": [], "categories": [], "annotations": []}
            for s in splits
        }
        ann_id_counter: dict[str, int] = dict.fromkeys(splits, 1)

        grouped = prepared_ldf.processed_df.group_by(
            ["file", "instance_id", "group_id"], maintain_order=True
        )
        copied_files: set[Path] = set()

        for key, entry in grouped:
            file_name, _instance_id, group_id = cast(tuple[str, int, Any], key)

            split = split_of_group(prepared_ldf, group_id)
            file_path = Path(str(file_name))

            image_id, width, height, new_name = self._get_or_register_image(
                file_path, split, annotation_splits
            )

            ann = {
                "id": ann_id_counter[split],
                "image_id": image_id,
                "category_id": None,
                "bbox": [],
                "area": 0,
                "iscrowd": 0,
            }
            has_valid_ann = False

            for row in entry.iter_rows(named=True):
                ann, row_has_ann = self._process_row(
                    row, split, annotation_splits, ann, width, height
                )
                has_valid_ann = has_valid_ann or row_has_ann

            img_size = file_path.stat().st_size
            if has_valid_ann:
                ann_size = sys.getsizeof(ann)
                annotation_splits = self._maybe_roll_partition(
                    annotation_splits, ann_size + img_size
                )

                annotation_splits[split]["annotations"].append(ann)
                ann_id_counter[split] += 1

            data_path = self._get_data_path(self.output_path, split, self.part)
            data_path.mkdir(parents=True, exist_ok=True)
            if file_path not in copied_files:
                copied_files.add(file_path)
                dest = data_path / new_name
                if dest != file_path:
                    dest.write_bytes(file_path.read_bytes())
                self.current_size += img_size

        self._dump_annotations(annotation_splits, self.output_path, self.part)

    def _maybe_roll_partition(
        self,
        annotation_splits: dict[str, dict[str, Any]],
        additional_size: int,
    ) -> dict[str, dict[str, Any]]:
        if (
            self.max_partition_size
            and self.part is not None
            and (self.current_size + additional_size) > self.max_partition_size
        ):
            self._dump_annotations(
                annotation_splits, self.output_path, self.part
            )
            self.current_size = 0
            self.part += 1
            return {
                s: {"images": [], "categories": [], "annotations": []}
                for s in self.get_split_names()
            }
        return annotation_splits

    def _get_or_register_image(
        self,
        path: Path,
        split: str,
        annotation_splits: dict[str, dict[str, Any]],
    ) -> tuple[int, int, int, str]:
        idx = self.image_indices.setdefault(path, len(self.image_indices))
        new_name = f"{idx}{path.suffix}"
        if str(path) in self.image_registry[split]:
            img_data = self.image_registry[split][str(path)]
            return (
                img_data["id"],
                img_data["width"],
                img_data["height"],
                new_name,
            )

        width, height = Image.open(path).size
        image_id = len(annotation_splits[split]["images"]) + 1
        img_entry = {
            "id": image_id,
            "file_name": new_name,
            "height": height,
            "width": width,
        }
        annotation_splits[split]["images"].append(img_entry)
        self.image_registry[split][str(path)] = img_entry
        return image_id, width, height, new_name

    def _process_row(
        self,
        row: dict[str, Any],
        split: str,
        annotation_splits: dict[str, dict[str, Any]],
        ann: dict[str, Any],
        w: int,
        h: int,
    ) -> tuple[dict[str, Any], bool]:
        ttype = row["task_type"]
        ann_str = row["annotation"]
        cname = row["class_name"]

        if cname and cname not in self.class_name_to_category_id[split]:
            cid = self.last_category_id[split]

            cat_entry = {"id": cid, "name": cname}

            if self.allow_keypoints:
                kp_labels, kp_skeleton = get_single_skeleton(
                    self.allow_keypoints, self.skeletons
                )
                if kp_labels:
                    cat_entry["keypoints"] = kp_labels
                    cat_entry["skeleton"] = kp_skeleton

            annotation_splits[split]["categories"].append(cat_entry)
            self.class_name_to_category_id[split][cname] = cid
            self.last_category_id[split] += 1

        if ttype == "classification" and cname:
            ann["category_id"] = self.class_name_to_category_id[split][cname]
            return ann, True

        if ann_str is None:
            return ann, False

        if ttype == "boundingbox":
            data = json.loads(ann_str)
            self._fill_bbox(ann, data, w, h, split, cname)
            return ann, True

        if ttype == "instance_segmentation":
            data = json.loads(ann_str)
            self._fill_instance_segmentation(ann, data, split, cname)
            return ann, True

        if ttype == "keypoints" and self.allow_keypoints:
            data = json.loads(ann_str)
            self._fill_keypoints(ann, data, w, h, split, cname)
            return ann, True

        return ann, False

    def _fill_bbox(
        self,
        ann: dict[str, Any],
        data: dict[str, Any],
        w: int,
        h: int,
        split: str,
        cname: str,
    ) -> None:
        ann["category_id"] = (
            ann.get("category_id")
            or self.class_name_to_category_id[split][cname]
        )
        x, y, bw, bh = (
            data.get("x", 0.0),
            data.get("y", 0.0),
            data.get("w", 0.0),
            data.get("h", 0.0),
        )
        px, py, pw, ph = x * w, y * h, bw * w, bh * h
        ann["bbox"] = [px, py, pw, ph]
        ann["area"] = pw * ph

    def _fill_instance_segmentation(
        self,
        ann: dict[str, Any],
        data: dict[str, Any],
        split: str,
        cname: str,
    ) -> None:
        ann["category_id"] = (
            ann.get("category_id")
            or self.class_name_to_category_id[split][cname]
        )
        H = int(data["height"])
        W = int(data["width"])
        counts = data["counts"]
        ann["segmentation"] = {"size": [H, W], "counts": counts}
        ann["iscrowd"] = 0 if ann.get("iscrowd") is None else ann["iscrowd"]

        if counts is not None:
            rle_runtime = {
                "size": [H, W],
                "counts": counts.encode("utf-8"),
            }
            area = float(maskUtils.area(rle_runtime))  # type: ignore[arg-type]
            bbox = maskUtils.toBbox(rle_runtime).tolist()  # type: ignore[arg-type]
            ann["area"] = area
            if not ann.get("bbox"):
                ann["bbox"] = bbox
            return

        ann["area"] = H * W
        ann["bbox"] = [0.0, 0.0, float(W), float(H)]

    def _fill_keypoints(
        self,
        ann: dict[str, Any],
        data: dict[str, Any],
        w: int,
        h: int,
        split: str,
        cname: str,
    ) -> None:
        ann["category_id"] = (
            ann.get("category_id")
            or self.class_name_to_category_id[split][cname]
        )

        raw_kps = data.get("keypoints", [])
        kps_px: list[float] = []
        xs: list[float] = []
        ys: list[float] = []
        visible_count = 0

        for triplet in raw_kps:
            x_norm, y_norm, v = triplet
            x_px = float(x_norm) * float(w)
            y_px = float(y_norm) * float(h)

            kps_px.extend([x_px, y_px, int(v)])

            if int(v) > 0:
                xs.append(x_px)
                ys.append(y_px)
                visible_count += 1

        ann["keypoints"] = kps_px
        ann["num_keypoints"] = visible_count

        if visible_count > 0:
            x_min = float(min(xs))
            y_min = float(min(ys))
            x_max = float(max(xs))
            y_max = float(max(ys))
            bw = max(0.0, x_max - x_min)
            bh = max(0.0, y_max - y_min)
            if not ann.get("bbox"):
                ann["bbox"] = [x_min, y_min, bw, bh]
            ann["area"] = bw * bh
        else:
            ann.setdefault("bbox", [0.0, 0.0, 0.0, 0.0])
            ann.setdefault("area", 0.0)

    def _dump_annotations(
        self,
        annotation_splits: dict[str, dict[str, Any]],
        output_path: Path,
        part: int | None = None,
    ) -> None:
        for split_key, split_data in annotation_splits.items():
            split_name = self.get_split_names().get(split_key, split_key)
            base = (
                output_path / f"{self.dataset_identifier}_part{part}"
                if part is not None
                else output_path / self.dataset_identifier
            )
            out_dir = base / split_name
            if self.format == COCOFormat.FIFTYONE:
                out_dir = out_dir / "labels"
            out_dir.mkdir(parents=True, exist_ok=True)
            (out_dir / "_annotations.coco.json").write_text(
                json.dumps(split_data, indent=2), encoding="utf-8"
            )

    def _get_data_path(
        self, output_path: Path, split: str, part: int | None = None
    ) -> Path:
        split_name = self.get_split_names().get(split, split)
        base = (
            output_path / f"{self.dataset_identifier}_part{part}"
            if part is not None
            else output_path / self.dataset_identifier
        )
        data_path = base / split_name
        if self.format == COCOFormat.FIFTYONE:
            data_path = data_path / "data"
        return data_path



from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from luxonis_ml.data.exporters.exporter_utils import PreparedLDF


class BaseExporter(ABC):
    def __init__(
        self,
        dataset_identifier: str,
        output_path: Path,
        max_partition_size_gb: float | None,
    ):
        self.dataset_identifier = dataset_identifier
        self.output_path = output_path
        self.image_indices = {}

        # attributes for intermediate saving
        self.max_partition_size_gb = max_partition_size_gb
        self.max_partition_size = (
            self.max_partition_size_gb * 1024**3
            if self.max_partition_size_gb
            else None
        )
        self.part = 0 if max_partition_size_gb else None
        self.current_size = 0

    @abstractmethod
    def export(self, prepared_ldf: "PreparedLDF") -> None:
        """Convert the prepared dataset into the exporter's format."""
        raise NotImplementedError

    @abstractmethod
    def _get_data_path(self, output_path: Path, split: str) -> Path:
        """Return the folder path to store data files for this split."""
        raise NotImplementedError

    @abstractmethod
    def _dump_annotations(
        self, annotations: dict, output_path: Path, part: int | None = None
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def supported_ann_types(self) -> list[str]:
        raise NotImplementedError
